
#pragma once
#include <cmath>
#include <dune/pdelab/common/function.hh>
#include <dune/pdelab/constraints/common/constraintsparameters.hh>

template<typename RF>
class DomainSize
{
public:
  const RF bottom;
  const RF top;
  const RF left;
  const RF right;
  const RF domainEps;

  DomainSize(RF b, RF t, RF l, RF r, RF eps):
  bottom(b), top(t), left(l), right(r), domainEps(eps)
  {
  }
};

template<typename DomainSize>
class LeftBottomBoundary : public Dune::PDELab::DirichletConstraintsParameters
{
private:
  const DomainSize ds;
public:
  LeftBottomBoundary(const DomainSize d) : ds(d)
  {
  }

  template<typename I>
  bool isDirichlet(const I & intersection, const Dune::FieldVector<typename I::ctype, I::mydimension> & coord) const
  {
    Dune::FieldVector<typename I::ctype, I::coorddimension>
      xg = intersection.geometry().global( coord );
    if (( xg[0] > ds.left + ds.domainEps) || ( xg[1]>ds.bottom + ds.domainEps))
    {
      return false;
    }
    return true;
  }
};


template<typename DomainSize>
class LeftBoundary : public Dune::PDELab::DirichletConstraintsParameters
{
private:
  const DomainSize ds;
public:
  LeftBoundary(const DomainSize d) : ds(d)
  {
  }

  template<typename I>
  bool isDirichlet(const I & intersection, const Dune::FieldVector<typename I::ctype, I::mydimension> & coord) const
  {
    Dune::FieldVector<typename I::ctype, I::coorddimension>
      xg = intersection.geometry().global( coord );
    if ( xg[0] > ds.domainEps)
    {
      return false;
    }
    return true;
  }
};

template<typename DomainSize>
class BottomBoundary : public Dune::PDELab::DirichletConstraintsParameters
{
private:
  const DomainSize ds;
public:
  BottomBoundary(const DomainSize d) : ds(d)
  {
  }

  template<typename I>
  bool isDirichlet(const I & intersection, const Dune::FieldVector<typename I::ctype, I::mydimension> & coord) const
  {
    Dune::FieldVector<typename I::ctype, I::coorddimension>
      xg = intersection.geometry().global( coord );
      if ( xg[1]>ds.bottom + ds.domainEps)
      {
        return false;
      }
      return true;
  }
};

template<typename DomainSize>
class LeftRightBoundary : public Dune::PDELab::DirichletConstraintsParameters
{
private:
  const DomainSize ds;
public:
  LeftRightBoundary(const DomainSize d) : ds(d)
  {
  }

  template<typename I>
  bool isDirichlet(const I & intersection, const Dune::FieldVector<typename I::ctype, I::mydimension> & coord) const
  {
    Dune::FieldVector<typename I::ctype, I::coorddimension>
      xg = intersection.geometry().global( coord );
      if (xg[1]>ds.bottom + ds.domainEps && xg[1]<ds.top - ds.domainEps)
      {
        return true;
      }
      return false;
  }
};

template<typename DomainSize>
class AllBoundary : public Dune::PDELab::DirichletConstraintsParameters
{
private:
  const DomainSize ds;
public:
  AllBoundary(const DomainSize d) : ds(d)
  {
  }

  template<typename I>
  bool isDirichlet(const I & intersection, const Dune::FieldVector<typename I::ctype, I::mydimension> & coord) const
  {
    return true;
  }
};

template<typename DomainSize>
class NoBoundary : public Dune::PDELab::DirichletConstraintsParameters
{
private:
  const DomainSize ds;
public:
  NoBoundary(const DomainSize d) : ds(d)
  {
  }

  template<typename I>
  bool isDirichlet(const I & intersection, const Dune::FieldVector<typename I::ctype, I::mydimension> & coord) const
  {
    return false;
  }
};
