#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include<iostream>
// dune includes
#include<dune/common/parallel/mpihelper.hh>
#include<dune/common/parametertreeparser.hh>
#include<dune/common/timer.hh>
//#if HAVE_UG
//#include<dune/grid/uggrid.hh>
//#endif
#if HAVE_DUNE_ALUGRID
#include<dune/alugrid/grid.hh>
#include<dune/alugrid/dgf.hh>
#include<dune/grid/io/file/dgfparser/dgfparser.hh>
#endif
#include <dune/grid/geometrygrid/grid.hh>
// pdelab includes
#include<dune/pdelab/finiteelementmap/pkfem.hh>
#include <dune/xt/grid/view/periodic.hh>
#include "3p_driver_monolithic.hh"
//#include "3p_driver.hh"

//===============================================================
// Main program with grid setup
//===============================================================

/*template<typename RF>
class ExampleThinstripTransform
      : public Dune :: AnalyticalCoordFunction< RF, 2, 2, ExampleThinstripTransform<RF> >
{
        typedef ExampleThinstripTransform<RF> This;
        typedef Dune :: AnalyticalCoordFunction< RF, 2, 2, This > Base;
      public:
        typedef typename Base :: DomainVector DomainVector;
        typedef typename Base :: RangeVector RangeVector;

        const RF YS;

        ExampleThinstripTransform(const RF YScalingFactor) : YS(YScalingFactor)
        {
        }

        void evaluate ( const DomainVector &x, RangeVector &y ) const
        {
          y[ 0 ] = x[ 0 ];
          y[ 1 ] = YS*x[ 1 ];
        }
};

template<typename RF>
class StrechingThinstripTransform
      : public Dune :: AnalyticalCoordFunction< RF, 2, 2, StrechingThinstripTransform<RF> >
{
        typedef StrechingThinstripTransform<RF> This;
        typedef Dune :: AnalyticalCoordFunction< RF, 2, 2, This > Base;
      public:
        typedef typename Base :: DomainVector DomainVector;
        typedef typename Base :: RangeVector RangeVector;

        const RF S;
        RF c;
        const RF q = 0.05;

        StrechingThinstripTransform(const RF Streching) : S(Streching)
        {
          c = q/S;
        }

        void evaluate ( const DomainVector &x, RangeVector &y ) const
        {
          y[ 1 ] = x[ 1 ];
          if( x[0] < c)
          {
            y[0] = x[0];
            return;
          }
          else if( x[0] > 1-c)
          {
            y[0] = x[0] + 1 - 1/S;
          }
          else
          {
            y[0] = (x[0]-c)*(S-2*q)/(1-2*q)+c;
          }
        }
};*/


int main(int argc, char** argv)
{
  try{
    // Maybe initialize Mpi
    Dune::MPIHelper&
      helper = Dune::MPIHelper::instance(argc, argv);
    if(Dune::MPIHelper::isFake)
      std::cout<< "This is a sequential program." << std::endl;
    else
      std::cout << "Parallel code run on "
                << helper.size() << " process(es)" << std::endl;

    // open ini file
    Dune::ParameterTree ptree;
    Dune::ParameterTreeParser ptreeparser;
    ptreeparser.readINITree("chns.ini",ptree);
    ptreeparser.readOptions(argc,argv,ptree);

    if(argc > 1)
    {
      std::cout << argv[1] << std::endl;
      int factorsize = std::stoi( argv[1] );
      ptree["output.filename"] = "output" + std::to_string(factorsize);
      ptree["domain.filename"] = "grids/tper"  + std::to_string(factorsize) + ".msh";
      ptree["domain.YScaling"] = std::to_string(1./float(factorsize));

      ptree["Time.dt"] = std::to_string(1./float(factorsize) * ptree.get("Time.dt",0.1));
      ptree["Time.dtMax"] = std::to_string(1./float(factorsize) * ptree.get("Time.dtMax",0.1));
      ptree["Time.dtMin"] = std::to_string(1./float(factorsize) * ptree.get("Time.dtMin",0.1));
    }

//    const int dim = ptree.get<int>("grid.dim");
    const int refinement = ptree.get<int>("domain.level");


    if (true/*dim==2*/)
        {

          typedef Dune::ALUGrid<2,2,Dune::simplex,Dune::conforming> Grid;
          //typedef Dune::UGGrid<2> Grid;
          typedef Grid::ctype DF;

          std::vector<int> boundary_index_map;
          std::vector<int> element_index_map;

          std::string filename = ptree.get("domain.filename",
                                           "turbtube2d.msh");

// Unstructured Grid
          Dune::GridFactory<Grid> factory;
          Dune::GmshReader<Grid>::read(factory,filename,true,true);
          Dune::GridFactory<Grid>::WorldMatrix matrix;
          std::cout << "a" << std::endl;
          matrix[0][0] = 1;     matrix[0][1] = 0;
          matrix[1][0] = 0;     matrix[1][1] = 1;
          std::cout << "b" << std::endl;
          Dune::GridFactory<Grid>::WorldVector shift;
          shift[0] = ptree.get("domain.len",1);
          shift[1] = 0;
          std::cout << "c" << std::endl;
          factory.insertFaceTransformation(matrix,shift );
          std::shared_ptr<Grid> gridp(factory.createGrid());
          std::cout << "d" << std::endl;





// Transforming big rectangle by shortening y-Direction

          // Dune::FieldVector<DF,2> lowerLeft(0);
          // Dune::FieldVector<DF,2> upperRight(1);
          // upperRight[1]=1./ptree.get("domain.OneOverYGridWidth", 2);
          // std::array<unsigned int,2> elements;
          // elements[0]=ptree.get("domain.CellsInYDirection", 32)*ptree.get("domain.OneOverYGridWidth", 2);
          // elements[1]=ptree.get("domain.CellsInYDirection", 32);

          //typedef Dune::GeometryGrid<Grid,ExampleThinstripTransform<double>> GGrid;
          //ExampleThinstripTransform<double>* tfp = new ExampleThinstripTransform<double>(ptree.get("domain.YScalingFactor",1.0));

// Transforming small rectangle by elongating in x-Direction
          // double A = ptree.get("domain.OneOverYGridWidth", 2);
          // double B = ptree.get("domain.YScalingFactor",1.0);
          // Dune::FieldVector<DF,2> lowerLeft(0);
          // Dune::FieldVector<DF,2> upperRight(0);
          // upperRight[0]=B;
          // upperRight[1]=B/A;
          // std::array<unsigned int,2> elements;
          // elements[0]=ptree.get("domain.CellsInYDirection", 32)*A;
          // elements[1]=ptree.get("domain.CellsInYDirection", 32);
          //
          // typedef Dune::GeometryGrid<Grid,StrechingThinstripTransform<double>> GGrid;
          // StrechingThinstripTransform<double>* tfp = new StrechingThinstripTransform<double>(1/B);
// End of possible Transformations


// Structured Grid for Transformations
          // std::shared_ptr<Grid> gridp(Dune::StructuredGridFactory<Grid>::createSimplexGrid(lowerLeft, upperRight, elements));
          // std::shared_ptr<GGrid> ggridp(new GGrid(gridp.get(),tfp));
          // typedef GGrid::LeafGridView LGV;

          typedef Grid::LeafGridView LGV;
          std::bitset<2> periodic_directions(ptree.get("domain.periodic","00"));
          typedef Dune::XT::Grid::PeriodicGridView<LGV> GV;
          GV gv(gridp->leafGridView(),periodic_directions);

          //typedef Grid::LeafGridView GV;
          //GV gv = gridp->leafGridView();

          std::cout << "OK" << std::endl;

          using ES = Dune::PDELab::AllEntitySet<GV>;
          ES es(gv);

          const int k = 2;
          typedef Dune::PDELab::PkLocalFiniteElementMap<ES,DF,double,k> V_FEM;
          typedef Dune::PDELab::PkLocalFiniteElementMap<ES,DF,double,k-1> P_FEM;
          const int l = 1;
          typedef Dune::PDELab::PkLocalFiniteElementMap<ES,DF,double,l> PHI_FEM;

          V_FEM vFem(es);
          P_FEM pFem(es);
          PHI_FEM phiFem(es);

          driver_3p(*gridp, gv, es,vFem,pFem,phiFem, ptree, helper);
        }
      }
  catch (Dune::Exception &e){
    std::cerr << "Dune reported error: " << e << std::endl;
    return 1;
  }
  catch (...){
    std::cerr << "Unknown exception thrown!" << std::endl;
    return 1;
  }
}
